package cz.muni.fi.pa165.core.company;

import cz.muni.fi.pa165.model.dto.common.Result;
import cz.muni.fi.pa165.model.dto.company.CompanyCreateDto;
import cz.muni.fi.pa165.model.dto.company.CompanyDto;
import cz.muni.fi.pa165.model.dto.company.CompanyUpdateDto;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.Parameter;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.media.Schema;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.tags.Tag;
import jakarta.validation.Valid;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;

import java.util.List;

@RestController
@RequestMapping("/api/company")
@Tag(name = "Company", description = "Device API")

public class CompanyController {
	private CompanyFacade companyFacade;

	@Autowired
	public CompanyController(CompanyFacade companyFacade) {
		this.companyFacade = companyFacade;
	}

	@Operation(summary = "Find all companies with pagination", responses = {
			@ApiResponse(responseCode = "200", description = "List of companies with pagination", content = {
					@Content(mediaType = MediaType.APPLICATION_JSON_VALUE, schema = @Schema(implementation = CompanyDto.class))
			})
	})
	@GetMapping("/pageable")
	public List<CompanyDto> findAllPageable(Pageable pageable) {
		return companyFacade.findAllPageable(pageable);
	}

	@Operation(summary = "Find all companies with pagination", responses = {
			@ApiResponse(responseCode = "200", description = "List of companies with pagination", content = {
					@Content(mediaType = MediaType.APPLICATION_JSON_VALUE, schema = @Schema(implementation = Result.class))
			})
	})
	@GetMapping()
	public Result<CompanyDto> findAll(
			@Parameter(description = "Page number of results to retrieve") @RequestParam int page) {
		return companyFacade.findAll(page);
	}

	@Operation(summary = "Find all companies", responses = {
			@ApiResponse(responseCode = "200", description = "List of companies", content = {
					@Content(mediaType = MediaType.APPLICATION_JSON_VALUE, schema = @Schema(implementation = CompanyDto.class))
			})
	})
	@GetMapping("/all")
	public List<CompanyDto> findAll() {
		return companyFacade.findAll();
	}

	@Operation(summary = "Find company by id", responses = {
			@ApiResponse(responseCode = "200", description = "Company found", content = {
					@Content(mediaType = MediaType.APPLICATION_JSON_VALUE, schema = @Schema(implementation = CompanyDto.class))
			}),
			@ApiResponse(responseCode = "404", description = "Company not found", content = @Content)
	})
	@GetMapping("/{id}")
	public CompanyDto findById(@PathVariable @Parameter(description = "The id of the company.") String id) {
		return companyFacade.findById(id);
	}

	@PostMapping
	@Operation(summary = "Create a new company")
	@ApiResponse(responseCode = "201", description = "The newly created company",
			content = @Content(mediaType = MediaType.APPLICATION_JSON_VALUE,
					schema = @Schema(implementation = CompanyDto.class)))
	@ApiResponse(responseCode = "400", description = "Invalid input", content = @Content)
	@ApiResponse(responseCode = "409", description = "Company with the same name already exists", content = @Content)
	public ResponseEntity<CompanyDto> create(@RequestBody @Valid CompanyCreateDto companyCreateDto) {
		return ResponseEntity.status(HttpStatus.CREATED).body(companyFacade.create(companyCreateDto));
	}

	@PutMapping("/{id}")
	@Operation(summary = "Update a company by ID")
	@ApiResponse(responseCode = "200", description = "The updated company",
			content = @Content(mediaType = MediaType.APPLICATION_JSON_VALUE,
					schema = @Schema(implementation = CompanyDto.class)))
	public CompanyDto updateById(@PathVariable @Parameter(description = "The id of the company.") String id,
								 @RequestBody @Valid CompanyUpdateDto companyUpdateDto) {
		return companyFacade.updateById(companyUpdateDto, id);
	}

	@DeleteMapping("/{id}")
	@Operation(summary = "Delete a company by ID")
	@ApiResponse(responseCode = "200", description = "The deleted company",
			content = @Content(mediaType = MediaType.APPLICATION_JSON_VALUE,
					schema = @Schema(implementation = CompanyDto.class)))
	public CompanyDto deleteById(@PathVariable @Parameter(description = "The id of the company.") String id) {
		return companyFacade.deleteById(id);
	}
}
