package cz.muni.fi.pa165.core.smartmeter;

import cz.muni.fi.pa165.model.dto.smartDevice.SmartMeterCreateDto;
import cz.muni.fi.pa165.model.dto.smartDevice.SmartMeterDto;
import cz.muni.fi.pa165.model.dto.smartDevice.SmartMeterUpdateDto;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.Parameter;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.tags.Tag;
import jakarta.validation.Valid;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Pageable;
import org.springframework.web.bind.annotation.CrossOrigin;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import java.util.List;

@RestController
@RequestMapping("/api/smart-meter")
@Tag(name = "smart-meter", description = "SmartMeter API")
public class SmartMeterController {

	private final SmartMeterFacade smartMeterFacade;

	@Autowired
	public SmartMeterController(SmartMeterFacade smartMeterFacade) {
		this.smartMeterFacade = smartMeterFacade;
	}

	@Operation(summary = "Find all smart meters (pageable)", description = "Returns a pageable list of all smart meters.")
	@GetMapping("/pageable")
	@ApiResponse(responseCode = "200", description = "Successfully retrieved the pageable list of smart meters.")
	public List<SmartMeterDto> findAllPageable(Pageable pageable) {
		return smartMeterFacade.findAllPageable(pageable);
	}

	@Operation(summary = "Find all smart meters", description = "Returns a list of all smart meters.")
	@GetMapping
	@ApiResponse(responseCode = "200", description = "Successfully retrieved the list of smart meters.")
	@CrossOrigin(origins = "*") // CORS headers needed for JavaScript clients
	public List<SmartMeterDto> findAll() {
		return smartMeterFacade.findAll();
	}

	@Operation(summary = "Find smart meter by ID", description = "Returns the smart meter with the specified ID.")
	@GetMapping("/{id}")
	@ApiResponse(responseCode = "200", description = "Successfully retrieved the smart meter.")
	@ApiResponse(responseCode = "404", description = "Smart meter not found.")
	public SmartMeterDto findById(@PathVariable @Parameter(description = "The ID of the smart meter to retrieve.") String id) {
		return smartMeterFacade.findById(id);
	}

	@Operation(summary = "Create smart meter", description = "Creates a new smart meter.")
	@PostMapping
	@ApiResponse(responseCode = "201", description = "Successfully created a new smart meter.")
	public SmartMeterDto create(@RequestBody @Valid SmartMeterCreateDto smartMeterCreateDto) {
		return smartMeterFacade.create(smartMeterCreateDto);
	}

	@Operation(summary = "Delete smart meter", description = "Deletes the smart meter with the specified ID.")
	@DeleteMapping("/{id}")
	@ApiResponse(responseCode = "200", description = "Successfully deleted the smart meter.")
	@ApiResponse(responseCode = "404", description = "Smart meter not found.")
	public SmartMeterDto deleteById(@PathVariable @Parameter(description = "The ID of the smart meter to delete.") String id) {
		return smartMeterFacade.deleteById(id);
	}

	@Operation(summary = "Update smart meter", description = "Updates the smart meter with the specified ID.")
	@PutMapping("/{id}")
	@ApiResponse(responseCode = "200", description = "Successfully updated the smart meter.")
	@ApiResponse(responseCode = "404", description = "Smart meter not found.")
	public SmartMeterDto updateById(@PathVariable @Parameter(description = "The ID of the smart meter to update.") String id,
									@RequestBody @Valid SmartMeterUpdateDto smartMeterUpdateDto) {
		return smartMeterFacade.updateById(smartMeterUpdateDto, id);
	}
}
